@@ -61,15 +61,17 @@ func newHighestTimestampMetric() *maxTimestamp {
 
 func TestSampleDelivery(t *testing.T) {
 	testcases := []struct {
-		name       string
-		samples    bool
-		exemplars  bool
-		histograms bool
+		name            string
+		samples         bool
+		exemplars       bool
+		histograms      bool
+		floatHistograms bool
 	}{
-		{samples: true, exemplars: false, histograms: false, name: "samples only"},
-		{samples: true, exemplars: true, histograms: true, name: "samples, exemplars, and histograms"},
-		{samples: false, exemplars: true, histograms: false, name: "exemplars only"},
-		{samples: false, exemplars: false, histograms: true, name: "histograms only"},
+		{samples: true, exemplars: false, histograms: false, floatHistograms: false, name: "samples only"},
+		{samples: true, exemplars: true, histograms: true, floatHistograms: true, name: "samples, exemplars, and histograms"},
+		{samples: false, exemplars: true, histograms: false, floatHistograms: false, name: "exemplars only"},
+		{samples: false, exemplars: false, histograms: true, floatHistograms: false, name: "histograms only"},
+		{samples: false, exemplars: false, histograms: false, floatHistograms: true, name: "float histograms only"},
 	}
 
 	// Let's create an even number of send batches so we don't run into the
@@ -101,10 +103,11 @@ func TestSampleDelivery(t *testing.T) {
 	for _, tc := range testcases {
 		t.Run(tc.name, func(t *testing.T) {
 			var (
-				series     []record.RefSeries
-				samples    []record.RefSample
-				exemplars  []record.RefExemplar
-				histograms []record.RefHistogramSample
+				series          []record.RefSeries
+				samples         []record.RefSample
+				exemplars       []record.RefExemplar
+				histograms      []record.RefHistogramSample
+				floatHistograms []record.RefFloatHistogramSample
 			)
 
 			// Generates same series in both cases.
@@ -115,7 +118,10 @@ func TestSampleDelivery(t *testing.T) {
 				exemplars, series = createExemplars(n, n)
 			}
 			if tc.histograms {
-				histograms, series = createHistograms(n, n)
+				histograms, _, series = createHistograms(n, n, false)
+			}
+			if tc.floatHistograms {
+				_, floatHistograms, series = createHistograms(n, n, true)
 			}
 
 			// Apply new config.
@@ -135,18 +141,22 @@ func TestSampleDelivery(t *testing.T) {
 			c.expectSamples(samples[:len(samples)/2], series)
 			c.expectExemplars(exemplars[:len(exemplars)/2], series)
 			c.expectHistograms(histograms[:len(histograms)/2], series)
+			c.expectFloatHistograms(floatHistograms[:len(floatHistograms)/2], series)
 			qm.Append(samples[:len(samples)/2])
 			qm.AppendExemplars(exemplars[:len(exemplars)/2])
 			qm.AppendHistograms(histograms[:len(histograms)/2])
+			qm.AppendFloatHistograms(floatHistograms[:len(floatHistograms)/2])
 			c.waitForExpectedData(t)
 
 			// Send second half of data.
 			c.expectSamples(samples[len(samples)/2:], series)
 			c.expectExemplars(exemplars[len(exemplars)/2:], series)
 			c.expectHistograms(histograms[len(histograms)/2:], series)
+			c.expectFloatHistograms(floatHistograms[len(floatHistograms)/2:], series)
 			qm.Append(samples[len(samples)/2:])
 			qm.AppendExemplars(exemplars[len(exemplars)/2:])
 			qm.AppendHistograms(histograms[len(histograms)/2:])
+			qm.AppendFloatHistograms(floatHistograms[len(floatHistograms)/2:])
 			c.waitForExpectedData(t)
 		})
 	}
@@ -161,7 +171,7 @@ func TestMetadataDelivery(t *testing.T) {
 	mcfg := config.DefaultMetadataConfig
 
 	metrics := newQueueManagerMetrics(nil, "", "")
-	m := NewQueueManager(metrics, nil, nil, nil, dir, newEWMARate(ewmaWeight, shardUpdateDuration), cfg, mcfg, nil, nil, c, defaultFlushDeadline, newPool(), newHighestTimestampMetric(), nil, false, false)
+	m := NewQueueManager(metrics, nil, nil, nil, dir, newEWMARate(ewmaWeight, shardUpdateDuration), cfg, mcfg, labels.EmptyLabels(), nil, c, defaultFlushDeadline, newPool(), newHighestTimestampMetric(), nil, false, false)
 	m.Start()
 	defer m.Stop()
 
@@ -352,7 +362,7 @@ func TestReshard(t *testing.T) {
 	c.waitForExpectedData(t)
 }
 
-func TestReshardRaceWithStop(t *testing.T) {
+func TestReshardRaceWithStop(*testing.T) {
 	c := NewTestWriteClient()
 	var m *QueueManager
 	h := sync.Mutex{}
@@ -539,6 +549,7 @@ func TestShouldReshard(t *testing.T) {
 func createTimeseries(numSamples, numSeries int, extraLabels ...labels.Label) ([]record.RefSample, []record.RefSeries) {
 	samples := make([]record.RefSample, 0, numSamples)
 	series := make([]record.RefSeries, 0, numSeries)
+	b := labels.ScratchBuilder{}
 	for i := 0; i < numSeries; i++ {
 		name := fmt.Sprintf("test_metric_%d", i)
 		for j := 0; j < numSamples; j++ {
@@ -548,9 +559,16 @@ func createTimeseries(numSamples, numSeries int, extraLabels ...labels.Label) ([
 				V:   float64(i),
 			})
 		}
+		// Create Labels that is name of series plus any extra labels supplied.
+		b.Reset()
+		b.Add(labels.MetricName, name)
+		for _, l := range extraLabels {
+			b.Add(l.Name, l.Value)
+		}
+		b.Sort()
 		series = append(series, record.RefSeries{
 			Ref:    chunks.HeadSeriesRef(i),
-			Labels: append(labels.Labels{{Name: "__name__", Value: name}}, extraLabels...),
+			Labels: b.Labels(),
 		})
 	}
 	return samples, series
@@ -578,35 +596,50 @@ func createExemplars(numExemplars, numSeries int) ([]record.RefExemplar, []recor
 	return exemplars, series
 }
 
-func createHistograms(numSamples, numSeries int) ([]record.RefHistogramSample, []record.RefSeries) {
+func createHistograms(numSamples, numSeries int, floatHistogram bool) ([]record.RefHistogramSample, []record.RefFloatHistogramSample, []record.RefSeries) {
 	histograms := make([]record.RefHistogramSample, 0, numSamples)
+	floatHistograms := make([]record.RefFloatHistogramSample, 0, numSamples)
 	series := make([]record.RefSeries, 0, numSeries)
 	for i := 0; i < numSeries; i++ {
 		name := fmt.Sprintf("test_metric_%d", i)
 		for j := 0; j < numSamples; j++ {
-			h := record.RefHistogramSample{
-				Ref: chunks.HeadSeriesRef(i),
-				T:   int64(j),
-				H: &histogram.Histogram{
-					Schema:          2,
-					ZeroThreshold:   1e-128,
-					ZeroCount:       0,
-					Count:           2,
-					Sum:             0,
-					PositiveSpans:   []histogram.Span{{Offset: 0, Length: 1}},
-					PositiveBuckets: []int64{int64(i) + 1},
-					NegativeSpans:   []histogram.Span{{Offset: 0, Length: 1}},
-					NegativeBuckets: []int64{int64(-i) - 1},
-				},
+			hist := &histogram.Histogram{
+				Schema:          2,
+				ZeroThreshold:   1e-128,
+				ZeroCount:       0,
+				Count:           2,
+				Sum:             0,
+				PositiveSpans:   []histogram.Span{{Offset: 0, Length: 1}},
+				PositiveBuckets: []int64{int64(i) + 1},
+				NegativeSpans:   []histogram.Span{{Offset: 0, Length: 1}},
+				NegativeBuckets: []int64{int64(-i) - 1},
+			}
+
+			if floatHistogram {
+				fh := record.RefFloatHistogramSample{
+					Ref: chunks.HeadSeriesRef(i),
+					T:   int64(j),
+					FH:  hist.ToFloat(),
+				}
+				floatHistograms = append(floatHistograms, fh)
+			} else {
+				h := record.RefHistogramSample{
+					Ref: chunks.HeadSeriesRef(i),
+					T:   int64(j),
+					H:   hist,
+				}
+				histograms = append(histograms, h)
 			}
-			histograms = append(histograms, h)
 		}
 		series = append(series, record.RefSeries{
 			Ref:    chunks.HeadSeriesRef(i),
-			Labels: labels.Labels{{Name: "__name__", Value: name}},
+			Labels: labels.FromStrings("__name__", name),
 		})
 	}
-	return histograms, series
+	if floatHistogram {
+		return nil, floatHistograms, series
+	}
+	return histograms, nil, series
 }
 
 func getSeriesNameFromRef(r record.RefSeries) string {
@@ -614,18 +647,20 @@ func getSeriesNameFromRef(r record.RefSeries) string {
 }
 
 type TestWriteClient struct {
-	receivedSamples    map[string][]prompb.Sample
-	expectedSamples    map[string][]prompb.Sample
-	receivedExemplars  map[string][]prompb.Exemplar
-	expectedExemplars  map[string][]prompb.Exemplar
-	receivedHistograms map[string][]prompb.Histogram
-	expectedHistograms map[string][]prompb.Histogram
-	receivedMetadata   map[string][]prompb.MetricMetadata
-	writesReceived     int
-	withWaitGroup      bool
-	wg                 sync.WaitGroup
-	mtx                sync.Mutex
-	buf                []byte
+	receivedSamples         map[string][]prompb.Sample
+	expectedSamples         map[string][]prompb.Sample
+	receivedExemplars       map[string][]prompb.Exemplar
+	expectedExemplars       map[string][]prompb.Exemplar
+	receivedHistograms      map[string][]prompb.Histogram
+	receivedFloatHistograms map[string][]prompb.Histogram
+	expectedHistograms      map[string][]prompb.Histogram
+	expectedFloatHistograms map[string][]prompb.Histogram
+	receivedMetadata        map[string][]prompb.MetricMetadata
+	writesReceived          int
+	withWaitGroup           bool
+	wg                      sync.WaitGroup
+	mtx                     sync.Mutex
+	buf                     []byte
 }
 
 func NewTestWriteClient() *TestWriteClient {
@@ -696,6 +731,23 @@ func (c *TestWriteClient) expectHistograms(hh []record.RefHistogramSample, serie
 	c.wg.Add(len(hh))
 }
 
+func (c *TestWriteClient) expectFloatHistograms(fhs []record.RefFloatHistogramSample, series []record.RefSeries) {
+	if !c.withWaitGroup {
+		return
+	}
+	c.mtx.Lock()
+	defer c.mtx.Unlock()
+
+	c.expectedFloatHistograms = map[string][]prompb.Histogram{}
+	c.receivedFloatHistograms = map[string][]prompb.Histogram{}
+
+	for _, fh := range fhs {
+		seriesName := getSeriesNameFromRef(series[fh.Ref])
+		c.expectedFloatHistograms[seriesName] = append(c.expectedFloatHistograms[seriesName], FloatHistogramToHistogramProto(fh.T, fh.FH))
+	}
+	c.wg.Add(len(fhs))
+}
+
 func (c *TestWriteClient) waitForExpectedData(tb testing.TB) {
 	if !c.withWaitGroup {
 		return
@@ -712,9 +764,12 @@ func (c *TestWriteClient) waitForExpectedData(tb testing.TB) {
 	for ts, expectedHistogram := range c.expectedHistograms {
 		require.Equal(tb, expectedHistogram, c.receivedHistograms[ts], ts)
 	}
+	for ts, expectedFloatHistogram := range c.expectedFloatHistograms {
+		require.Equal(tb, expectedFloatHistogram, c.receivedFloatHistograms[ts], ts)
+	}
 }
 
-func (c *TestWriteClient) Store(_ context.Context, req []byte) error {
+func (c *TestWriteClient) Store(_ context.Context, req []byte, _ int) error {
 	c.mtx.Lock()
 	defer c.mtx.Unlock()
 	// nil buffers are ok for snappy, ignore cast error.
@@ -747,7 +802,12 @@ func (c *TestWriteClient) Store(_ context.Context, req []byte) error {
 
 		for _, histogram := range ts.Histograms {
 			count++
-			c.receivedHistograms[seriesName] = append(c.receivedHistograms[seriesName], histogram)
+			if histogram.IsFloatHistogram() {
+				c.receivedFloatHistograms[seriesName] = append(c.receivedFloatHistograms[seriesName], histogram)
+			} else {
+				c.receivedHistograms[seriesName] = append(c.receivedHistograms[seriesName], histogram)
+			}
+
 		}
 	}
 	if c.withWaitGroup {
@@ -783,7 +843,7 @@ func NewTestBlockedWriteClient() *TestBlockingWriteClient {
 	return &TestBlockingWriteClient{}
 }
 
-func (c *TestBlockingWriteClient) Store(ctx context.Context, _ []byte) error {
+func (c *TestBlockingWriteClient) Store(ctx context.Context, _ []byte, _ int) error {
 	c.numCalls.Inc()
 	<-ctx.Done()
 	return nil
@@ -804,10 +864,10 @@ func (c *TestBlockingWriteClient) Endpoint() string {
 // For benchmarking the send and not the receive side.
 type NopWriteClient struct{}
 
-func NewNopWriteClient() *NopWriteClient                            { return &NopWriteClient{} }
-func (c *NopWriteClient) Store(_ context.Context, req []byte) error { return nil }
-func (c *NopWriteClient) Name() string                              { return "nopwriteclient" }
-func (c *NopWriteClient) Endpoint() string                          { return "http://test-remote.com/1234" }
+func NewNopWriteClient() *NopWriteClient                           { return &NopWriteClient{} }
+func (c *NopWriteClient) Store(context.Context, []byte, int) error { return nil }
+func (c *NopWriteClient) Name() string                             { return "nopwriteclient" }
+func (c *NopWriteClient) Endpoint() string                         { return "http://test-remote.com/1234" }
 
 func BenchmarkSampleSend(b *testing.B) {
 	// Send one sample per series, which is the typical remote_write case
@@ -815,7 +875,7 @@ func BenchmarkSampleSend(b *testing.B) {
 	const numSeries = 10000
 
 	// Extra labels to make a more realistic workload - taken from Kubernetes' embedded cAdvisor metrics.
-	extraLabels := labels.Labels{
+	extraLabels := []labels.Label{
 		{Name: "kubernetes_io_arch", Value: "amd64"},
 		{Name: "kubernetes_io_instance_type", Value: "c3.somesize"},
 		{Name: "kubernetes_io_os", Value: "linux"},
@@ -902,56 +962,63 @@ func BenchmarkStartup(b *testing.B) {
 func TestProcessExternalLabels(t *testing.T) {
 	for _, tc := range []struct {
 		labels         labels.Labels
-		externalLabels labels.Labels
+		externalLabels []labels.Label
 		expected       labels.Labels
 	}{
 		// Test adding labels at the end.
 		{
-			labels:         labels.Labels{{Name: "a", Value: "b"}},
-			externalLabels: labels.Labels{{Name: "c", Value: "d"}},
-			expected:       labels.Labels{{Name: "a", Value: "b"}, {Name: "c", Value: "d"}},
+			labels:         labels.FromStrings("a", "b"),
+			externalLabels: []labels.Label{{Name: "c", Value: "d"}},
+			expected:       labels.FromStrings("a", "b", "c", "d"),
 		},
 
 		// Test adding labels at the beginning.
 		{
-			labels:         labels.Labels{{Name: "c", Value: "d"}},
-			externalLabels: labels.Labels{{Name: "a", Value: "b"}},
-			expected:       labels.Labels{{Name: "a", Value: "b"}, {Name: "c", Value: "d"}},
+			labels:         labels.FromStrings("c", "d"),
+			externalLabels: []labels.Label{{Name: "a", Value: "b"}},
+			expected:       labels.FromStrings("a", "b", "c", "d"),
 		},
 
 		// Test we don't override existing labels.
 		{
-			labels:         labels.Labels{{Name: "a", Value: "b"}},
-			externalLabels: labels.Labels{{Name: "a", Value: "c"}},
-			expected:       labels.Labels{{Name: "a", Value: "b"}},
+			labels:         labels.FromStrings("a", "b"),
+			externalLabels: []labels.Label{{Name: "a", Value: "c"}},
+			expected:       labels.FromStrings("a", "b"),
 		},
 
 		// Test empty externalLabels.
 		{
-			labels:         labels.Labels{{Name: "a", Value: "b"}},
-			externalLabels: labels.Labels{},
-			expected:       labels.Labels{{Name: "a", Value: "b"}},
+			labels:         labels.FromStrings("a", "b"),
+			externalLabels: []labels.Label{},
+			expected:       labels.FromStrings("a", "b"),
 		},
 
 		// Test empty labels.
 		{
-			labels:         labels.Labels{},
-			externalLabels: labels.Labels{{Name: "a", Value: "b"}},
-			expected:       labels.Labels{{Name: "a", Value: "b"}},
+			labels:         labels.EmptyLabels(),
+			externalLabels: []labels.Label{{Name: "a", Value: "b"}},
+			expected:       labels.FromStrings("a", "b"),
 		},
 
 		// Test labels is longer than externalLabels.
 		{
-			labels:         labels.Labels{{Name: "a", Value: "b"}, {Name: "c", Value: "d"}},
-			externalLabels: labels.Labels{{Name: "e", Value: "f"}},
-			expected:       labels.Labels{{Name: "a", Value: "b"}, {Name: "c", Value: "d"}, {Name: "e", Value: "f"}},
+			labels:         labels.FromStrings("a", "b", "c", "d"),
+			externalLabels: []labels.Label{{Name: "e", Value: "f"}},
+			expected:       labels.FromStrings("a", "b", "c", "d", "e", "f"),
 		},
 
 		// Test externalLabels is longer than labels.
 		{
-			labels:         labels.Labels{{Name: "c", Value: "d"}},
-			externalLabels: labels.Labels{{Name: "a", Value: "b"}, {Name: "e", Value: "f"}},
-			expected:       labels.Labels{{Name: "a", Value: "b"}, {Name: "c", Value: "d"}, {Name: "e", Value: "f"}},
+			labels:         labels.FromStrings("c", "d"),
+			externalLabels: []labels.Label{{Name: "a", Value: "b"}, {Name: "e", Value: "f"}},
+			expected:       labels.FromStrings("a", "b", "c", "d", "e", "f"),
+		},
+
+		// Adding with and without clashing labels.
+		{
+			labels:         labels.FromStrings("a", "b", "c", "d"),
+			externalLabels: []labels.Label{{Name: "a", Value: "xxx"}, {Name: "c", Value: "yyy"}, {Name: "e", Value: "f"}},
+			expected:       labels.FromStrings("a", "b", "c", "d", "e", "f"),
 		},
 	} {
 		require.Equal(t, tc.expected, processExternalLabels(tc.labels, tc.externalLabels))
