@@ -17,6 +17,7 @@ import (
 	"fmt"
 	"net/http"
 	"sort"
+	"strings"
 
 	"github.com/go-kit/log/level"
 	"github.com/gogo/protobuf/proto"
@@ -25,7 +26,9 @@ import (
 	dto "github.com/prometheus/client_model/go"
 	"github.com/prometheus/common/expfmt"
 	"github.com/prometheus/common/model"
+	"golang.org/x/exp/slices"
 
+	"github.com/prometheus/prometheus/model/histogram"
 	"github.com/prometheus/prometheus/model/labels"
 	"github.com/prometheus/prometheus/model/timestamp"
 	"github.com/prometheus/prometheus/model/value"
@@ -55,6 +58,8 @@ func (h *Handler) federation(w http.ResponseWriter, req *http.Request) {
 	h.mtx.RLock()
 	defer h.mtx.RUnlock()
 
+	ctx := req.Context()
+
 	if err := req.ParseForm(); err != nil {
 		http.Error(w, fmt.Sprintf("error parsing form values: %v", err), http.StatusBadRequest)
 		return
@@ -78,7 +83,7 @@ func (h *Handler) federation(w http.ResponseWriter, req *http.Request) {
 	)
 	w.Header().Set("Content-Type", string(format))
 
-	q, err := h.localStorage.Querier(req.Context(), mint, maxt)
+	q, err := h.localStorage.Querier(mint, maxt)
 	if err != nil {
 		federationErrors.Inc()
 		if errors.Cause(err) == tsdb.ErrNotReady {
@@ -96,44 +101,63 @@ func (h *Handler) federation(w http.ResponseWriter, req *http.Request) {
 
 	var sets []storage.SeriesSet
 	for _, mset := range matcherSets {
-		s := q.Select(true, hints, mset...)
+		s := q.Select(ctx, true, hints, mset...)
 		sets = append(sets, s)
 	}
 
 	set := storage.NewMergeSeriesSet(sets, storage.ChainedSeriesMerge)
 	it := storage.NewBuffer(int64(h.lookbackDelta / 1e6))
+	var chkIter chunkenc.Iterator
+Loop:
 	for set.Next() {
 		s := set.At()
 
 		// TODO(fabxc): allow fast path for most recent sample either
 		// in the storage itself or caching layer in Prometheus.
-		it.Reset(s.Iterator())
-
-		var t int64
-		var v float64
-		var ok bool
+		chkIter = s.Iterator(chkIter)
+		it.Reset(chkIter)
 
+		var (
+			t  int64
+			f  float64
+			fh *histogram.FloatHistogram
+		)
 		valueType := it.Seek(maxt)
-		if valueType == chunkenc.ValFloat {
-			t, v = it.At()
-		} else {
-			// TODO(beorn7): Handle histograms.
-			t, v, _, ok = it.PeekBack(1)
+		switch valueType {
+		case chunkenc.ValFloat:
+			t, f = it.At()
+		case chunkenc.ValFloatHistogram, chunkenc.ValHistogram:
+			t, fh = it.AtFloatHistogram()
+		default:
+			sample, ok := it.PeekBack(1)
 			if !ok {
-				continue
+				continue Loop
+			}
+			t = sample.T()
+			switch sample.Type() {
+			case chunkenc.ValFloat:
+				f = sample.F()
+			case chunkenc.ValHistogram:
+				fh = sample.H().ToFloat()
+			case chunkenc.ValFloatHistogram:
+				fh = sample.FH()
+			default:
+				continue Loop
 			}
 		}
 		// The exposition formats do not support stale markers, so drop them. This
 		// is good enough for staleness handling of federated data, as the
 		// interval-based limits on staleness will do the right thing for supported
 		// use cases (which is to say federating aggregated time series).
-		if value.IsStaleNaN(v) {
+		if value.IsStaleNaN(f) || (fh != nil && value.IsStaleNaN(fh.Sum)) {
 			continue
 		}
 
 		vec = append(vec, promql.Sample{
 			Metric: s.Labels(),
-			Point:  promql.Point{T: t, V: v},
+			T:      t,
+			F:      f,
+			H:      fh,
 		})
 	}
 	if ws := set.Warnings(); len(ws) > 0 {
@@ -146,7 +170,11 @@ func (h *Handler) federation(w http.ResponseWriter, req *http.Request) {
 		return
 	}
 
-	sort.Sort(byName(vec))
+	slices.SortFunc(vec, func(a, b promql.Sample) int {
+		ni := a.Metric.Get(labels.MetricName)
+		nj := b.Metric.Get(labels.MetricName)
+		return strings.Compare(ni, nj)
+	})
 
 	externalLabels := h.config.GlobalConfig.ExternalLabels.Map()
 	if _, ok := externalLabels[model.InstanceLabel]; !ok {
@@ -159,44 +187,63 @@ func (h *Handler) federation(w http.ResponseWriter, req *http.Request) {
 	sort.Strings(externalLabelNames)
 
 	var (
-		lastMetricName string
-		protMetricFam  *dto.MetricFamily
+		lastMetricName                          string
+		lastWasHistogram, lastHistogramWasGauge bool
+		protMetricFam                           *dto.MetricFamily
 	)
 	for _, s := range vec {
-		nameSeen := false
-		globalUsed := map[string]struct{}{}
-		protMetric := &dto.Metric{
-			Untyped: &dto.Untyped{},
+		isHistogram := s.H != nil
+		if isHistogram &&
+			format != expfmt.FmtProtoDelim && format != expfmt.FmtProtoText && format != expfmt.FmtProtoCompact {
+			// Can't serve the native histogram.
+			// TODO(codesome): Serve them when other protocols get the native histogram support.
+			continue
 		}
 
-		for _, l := range s.Metric {
+		nameSeen := false
+		globalUsed := map[string]struct{}{}
+		protMetric := &dto.Metric{}
+
+		err := s.Metric.Validate(func(l labels.Label) error {
 			if l.Value == "" {
 				// No value means unset. Never consider those labels.
 				// This is also important to protect against nameless metrics.
-				continue
+				return nil
 			}
 			if l.Name == labels.MetricName {
 				nameSeen = true
-				if l.Value == lastMetricName {
-					// We already have the name in the current MetricFamily,
-					// and we ignore nameless metrics.
-					continue
+				if l.Value == lastMetricName && // We already have the name in the current MetricFamily, and we ignore nameless metrics.
+					lastWasHistogram == isHistogram && // The sample type matches (float vs histogram).
+					// If it was a histogram, the histogram type (counter vs gauge) also matches.
+					(!isHistogram || lastHistogramWasGauge == (s.H.CounterResetHint == histogram.GaugeType)) {
+					return nil
 				}
+
+				// Since we now check for the sample type and type of histogram above, we will end up
+				// creating multiple metric families for the same metric name. This would technically be
+				// an invalid exposition. But since the consumer of this is Prometheus, and Prometheus can
+				// parse it fine, we allow it and bend the rules to make federation possible in those cases.
+
 				// Need to start a new MetricFamily. Ship off the old one (if any) before
 				// creating the new one.
 				if protMetricFam != nil {
 					if err := enc.Encode(protMetricFam); err != nil {
-						federationErrors.Inc()
-						level.Error(h.logger).Log("msg", "federation failed", "err", err)
-						return
+						return err
 					}
 				}
 				protMetricFam = &dto.MetricFamily{
 					Type: dto.MetricType_UNTYPED.Enum(),
 					Name: proto.String(l.Value),
 				}
+				if isHistogram {
+					if s.H.CounterResetHint == histogram.GaugeType {
+						protMetricFam.Type = dto.MetricType_GAUGE_HISTOGRAM.Enum()
+					} else {
+						protMetricFam.Type = dto.MetricType_HISTOGRAM.Enum()
+					}
+				}
 				lastMetricName = l.Value
-				continue
+				return nil
 			}
 			protMetric.Label = append(protMetric.Label, &dto.LabelPair{
 				Name:  proto.String(l.Name),
@@ -205,6 +252,12 @@ func (h *Handler) federation(w http.ResponseWriter, req *http.Request) {
 			if _, ok := externalLabels[l.Name]; ok {
 				globalUsed[l.Name] = struct{}{}
 			}
+			return nil
+		})
+		if err != nil {
+			federationErrors.Inc()
+			level.Error(h.logger).Log("msg", "federation failed", "err", err)
+			return
 		}
 		if !nameSeen {
 			level.Warn(h.logger).Log("msg", "Ignoring nameless metric during federation", "metric", s.Metric)
@@ -222,9 +275,42 @@ func (h *Handler) federation(w http.ResponseWriter, req *http.Request) {
 		}
 
 		protMetric.TimestampMs = proto.Int64(s.T)
-		protMetric.Untyped.Value = proto.Float64(s.V)
-		// TODO(beorn7): Handle histograms.
-
+		if !isHistogram {
+			lastHistogramWasGauge = false
+			protMetric.Untyped = &dto.Untyped{
+				Value: proto.Float64(s.F),
+			}
+		} else {
+			lastHistogramWasGauge = s.H.CounterResetHint == histogram.GaugeType
+			protMetric.Histogram = &dto.Histogram{
+				SampleCountFloat: proto.Float64(s.H.Count),
+				SampleSum:        proto.Float64(s.H.Sum),
+				Schema:           proto.Int32(s.H.Schema),
+				ZeroThreshold:    proto.Float64(s.H.ZeroThreshold),
+				ZeroCountFloat:   proto.Float64(s.H.ZeroCount),
+				NegativeCount:    s.H.NegativeBuckets,
+				PositiveCount:    s.H.PositiveBuckets,
+			}
+			if len(s.H.PositiveSpans) > 0 {
+				protMetric.Histogram.PositiveSpan = make([]*dto.BucketSpan, len(s.H.PositiveSpans))
+				for i, sp := range s.H.PositiveSpans {
+					protMetric.Histogram.PositiveSpan[i] = &dto.BucketSpan{
+						Offset: proto.Int32(sp.Offset),
+						Length: proto.Uint32(sp.Length),
+					}
+				}
+			}
+			if len(s.H.NegativeSpans) > 0 {
+				protMetric.Histogram.NegativeSpan = make([]*dto.BucketSpan, len(s.H.NegativeSpans))
+				for i, sp := range s.H.NegativeSpans {
+					protMetric.Histogram.NegativeSpan[i] = &dto.BucketSpan{
+						Offset: proto.Int32(sp.Offset),
+						Length: proto.Uint32(sp.Length),
+					}
+				}
+			}
+		}
+		lastWasHistogram = isHistogram
 		protMetricFam.Metric = append(protMetricFam.Metric, protMetric)
 	}
 	// Still have to ship off the last MetricFamily, if any.
@@ -235,15 +321,3 @@ func (h *Handler) federation(w http.ResponseWriter, req *http.Request) {
 		}
 	}
 }
-
-// byName makes a model.Vector sortable by metric name.
-type byName promql.Vector
-
-func (vec byName) Len() int      { return len(vec) }
-func (vec byName) Swap(i, j int) { vec[i], vec[j] = vec[j], vec[i] }
-
-func (vec byName) Less(i, j int) bool {
-	ni := vec[i].Metric.Get(labels.MetricName)
-	nj := vec[j].Metric.Get(labels.MetricName)
-	return ni < nj
-}
